# 
# Copyright (C) 2001, 2002 Jeff Dike (jdike@karaya.com)
# Licensed under the GPL
#

package UML;

use UML::MConsole;
use UML::KConfig;
use Expect;
use Socket;
use IO::File;
use strict;

my $cow_index = 0;

sub get_cow {
    while(-e "cow$cow_index"){
	$cow_index++;
    }
    my $cow = "cow$cow_index";
    $cow_index++;
    return($cow);
}

sub ubd_args {
    my $me = shift;
    my $args = "";

    foreach my $i (0 .. $#{$me->{ubd}}){
	my $ubd = $me->{ubd}->[$i];
	!defined($ubd->[1]) and next;
	$args .= "ubd$i=";
	if(defined($me->{do_cow})){
	    $ubd->[0] = get_cow();
	    $args .= $ubd->[0] . ",";
	}
	$args .= $ubd->[1] . " ";
    }
    return($args);
}

sub check_ubd {
    my $uml = shift;
    my $ubd = shift;

    ref($ubd) ne "HASH" and die "'ubd' requires a hash of devices to files";
    foreach my $device (keys(%$ubd)){
	$device =~ /\D/ and die "$device isn't a valid device number";
	$uml->{ubd}->[$device] = [ undef, $ubd->{$device} ];
	$uml->{args}->{ubd} = \&ubd_args;
    }
}

sub check_cow {
    my $uml = shift;
    my $do_cow = shift;

    defined($do_cow) and $uml->{do_cow} = 1;
}

sub do_debug {
    return("debug ");
}

sub check_debug {
    my $uml = shift;
    my $debug = shift;

    !$debug and return;
    $uml->{args}->{debug} = \&do_debug;
}

my %args = (
	    "ubd" => \&check_ubd,
	    "cow_ubds" => \&check_cow,
	    "debug" => \&check_debug,
	    "mem" => 1,
	    "con\\d*" => 1,
	    "ssl\\d*" => 1,
	    "umid" => 1,
	    "iomem" => 1,
	    "initrd" => 1,
	    "root" => 1
);

sub arg {
    my $uml = shift;
    my $arg = shift;
    my $val = shift;

    foreach my $good (keys(%args)){
	if($arg =~ /^$good$/){
	    if(ref($args{$arg}) eq 'CODE'){
		$args{$arg}->($uml, $val);
	    }
	    else {
		$uml->{args}->{$arg} = $val;
	    }
	    return;
	}
    }
    die "Unknown argument : $arg";
}

sub args {
    my $me = shift;

    while(@_){
	my $arg = shift;
	my $val = shift;
	if($arg eq "kernel"){
	    $me->{kernel} = $val;
	}
	elsif($arg eq "login_prompt"){
	    $me->{login_prompt} = $val;
	}
	elsif($arg eq "login"){
	    $me->{login} = $val;
	}
	elsif($arg eq "password_prompt"){
	    $me->{password_prompt} = $val;
	}
	elsif($arg eq "password"){
	    $me->{password} = $val;
	}
	elsif($arg eq "prompt"){
	    $me->{prompt} = $val;
	}
	elsif($arg eq "halt"){
	    $me->{halt} = $val;
	}
	else {
	    arg($me, $arg, $val);
	}
    }    
}

sub new {
    my $class = shift;
    my $config = shift;

    my $me = { kernel => undef, 
	       args => { },
	       halt => "halt",
	       expect_handle => undef,
	       state => "down",
	       ubd => [ (undef) x 8 ],
	       mc_notify => undef };

    !defined(socket(SOCK, AF_UNIX, SOCK_DGRAM, 0)) and 
	die "UML::new - socket failed : $!\n";
    $me->{mc_notify} = \*SOCK;

# XXX the -u causes mktemp to unlink the file - this is needed because
# we don't want a normal file, we want a socket.
    my $sock = `mktemp -u /tmp/mc_notify.XXXXXX`;
    chomp $sock;
    $? != 0 and die "UML::new - mktemp failed with status $?\n";

    !defined(bind($me->{mc_notify}, sockaddr_un($sock))) and 
	die "UML::new - bind failed : $!\n";

    $me->{args}->{mconsole} = "notify:$sock";
    
    bless($me, $class);
    return $me;
}

sub expect {
    my $me = shift;
    my $handle = shift;
    my $time_left = defined($me->{timer}) ? $me->{timer}->() : undef;

    my ($pos, $err, undef, $output, undef) = 
	$handle->expect($time_left, "Kernel panic:", @_);
    $pos == 1 and die "kernel paniced : '$output'\n";
    return($pos - 1);
}

sub change_password {
    my $me = shift;
    my $handle = shift;
    my $password = shift;
    my $prompt = shift;

    expect($me, $handle, "(current) UNIX password:");
    $handle->print("$password\n");
    expect($me, $handle, "Enter new UNIX password:");
    $handle->print("Ro0TrO0t\n");
    expect($me, $handle, "Retype new UNIX password:");
    $handle->print("Ro0TrO0t\n");
    expect($me, $handle, "-re", $prompt);
    $handle->print("passwd\n");
    expect($me, $handle, "Enter new UNIX password:");
    $handle->print("root\n");
    expect($me, $handle, "Retype new UNIX password:");
    $handle->print("root\n");
    expect($me, $handle, "-re", $prompt);
}

sub get_mconsole {
    my $socket = shift;
    my $file = "";

    !defined(recv($socket, $file, 4096, 0)) and 
	die "UML::get_mconsole - recv failed : $!";
    return(UML::MConsole->new($file));
}

my @mand = ("kernel", "login_prompt", "login", "password_prompt",
	    "password", "prompt");

sub boot {
    my $me = shift;
    my $log_file = shift;
    my $log;

    if($me->{state} ne "down"){
	warn "UML::boot : already booted";
	return;
    }
    for my $mand (@mand){
	!defined($me->{$mand}) and die "'$mand' must be defined";
    }

    my $cmd = "$me->{kernel} ";
    foreach my $arg (keys(%{$me->{args}})){
	my $val = $me->{args}->{$arg};
	if(ref($val) eq "CODE"){
	    $cmd .= $val->($me);
	}
	else {
	    $cmd .= "$arg=$val ";
	}
    }

    $me->{expect_handle} = Expect->spawn($cmd);
    !defined($me->{expect_handle}) and 
	die "Running UML ('$cmd') failed : $!\n";

    if(defined($log_file)){
	$log = $me->open_log($log_file);
	$me->{expect_handle}->log_stdout(0);
    }

    my $sh = Expect->exp_init($me->{mc_notify});
    while(1){
	my @active = Expect::test_handles(undef, $sh, $me->{expect_handle});
	if(($#active == 1) || ($active[0] == 0)){
	    $me->{mconsole} = get_mconsole($me->{mc_notify});
	    last;
	}
	$me->{expect_handle}->expect(0, "$me->{login_prompt}");
    }

    expect($me, $me->{expect_handle}, "$me->{login_prompt}");
    $me->{expect_handle}->print("$me->{login}\n");
    expect($me, $me->{expect_handle}, "$me->{password_prompt}");
    $me->{expect_handle}->print("$me->{password}\n");
    my $match = expect($me, $me->{expect_handle}, "-re", "$me->{prompt}", 
		       "Changing password");
    $match == 2 and 
	change_password($me, $me->{expect_handle}, $me->{password});
    $me->{state} = "up";
    return($log);
}

sub kill {
    my $me = shift;

    defined($me->{mconsole}) and $me->{mconsole}->halt();
    my $umid = $me->{args}->{umid};
    if(defined($umid) && (-d $umid) && (-f "$umid/pid")){
	my $pid = `cat "$umid/pid"`;
	chomp $pid;
	kill "SIGINT", $pid;
    }
    $me->{expect_handle}->hard_close();
}

sub command {
    my $me = shift;
    my $cmd = shift;
    my $interactor = shift;
    my @expects = ( "-re", "$me->{prompt}", "Kernel panic" );

    my $time_left = defined($me->{timer}) ? $me->{timer}->() : undef;
    $me->{expect_handle}->print("$cmd\n");

    my ($pos, $err, $output);
    $pos = $me->{expect_handle}->expect($time_left, "\n");
    $pos != 1 and die "UML::command - Failed to see newline after command\n";

    if(defined($interactor)){
	($pos, $err, undef, $output, undef) = 
	    $interactor->($me->{expect_handle}, $time_left, @expects);
    }
    else {
	($pos, $err, undef, $output, undef) = 
	    $me->{expect_handle}->expect($time_left, @expects);
    }

    $me->{expect_handle}->clear_accum();
    if(defined($err)){
	if($err == 1){
	    $me->{timer} = sub { return(5) };
	    $me->command(chr(3));
	    die "Timed out\n";
	}
	die "Expect error : $err\n";
    }
    elsif($pos == 1){
	$output =~ /(.*\S+)\s*$/s and $output = $1;
	return($output);
    }
    else {
	$me->kill();
	$me->{state} = "down";
	die "Kernel panic : '$output'\n";
    }
}

sub open_log {
    my $me = shift;
    my $log = shift;

    $me->{expect_handle}->log_file($log);
}

sub close_log {
    my $me = shift;
    my $log = shift;
    my @logs = $me->{expect_handle}->set_group();

    foreach my $i (0..$#logs){
	if($logs[$i] == $log){
	    splice @logs, $i, 1;
	    $log->hard_close();
	}
    }
    if(!@logs){
	my $fh = new IO::File "> /dev/null";
	push @logs, Expect->exp_init(\*$fh);
    }
    $me->{expect_handle}->set_group(@logs);
}

sub halt {
    my $me = shift;

    $me->{state} eq "down" and return;
    $me->{expect_handle}->print("$me->{halt}\n");
    my $time_left = defined($me->{timer}) ? $me->{timer}->() : undef;
    $me->{expect_handle}->expect($time_left);
    $me->{state} = "down";
    foreach my $ubd (@{$me->{ubd}}){
	my $cow = $ubd->[0];
	if(defined($cow)){
	    unlink $cow || warn "Couldn't unlink COW file '$cow' : $!\n";
	}
    }
}

sub timer {
    my $me = shift;
    my $timer = shift;

    $me->{timer} = $timer;
}

sub add {
    my $me = shift;
    my $file = shift;
    my %options = @_;

    !defined($me->{mconsole}) and die "UML::add - No mconsole available\n";
    ! -f $file and die "UML::add - '$file' not a plain file\n";

    my $arg = "$file";
    my $cow = undef;
    if((!defined($options{cow}) || ($options{cow} != 0)) && 
       defined($me->{do_cow})){
	$cow = get_cow();
    }

    my $disks = $me->{ubd};
    foreach my $i (0 .. $#$disks){
	defined($disks->[$i]) and next;
	$me->{ubd}->[$i] = [ $cow, $file ];
	defined($cow) and $arg = "$cow,$arg";
	$me->{mconsole}->config("ubd$i", $arg);
	return($i);
    }
    die "UML::add : No available devices\n";
}

sub remove {
    my $me = shift;
    my $dev = shift;

    !defined($me->{mconsole}) and die "UML::remove - No mconsole available\n";
    $me->{mconsole}->remove("ubd$dev");
    my $ubd = $me->{ubd}->[$dev];
    my ($cow, $back) = @$ubd;
    if(defined($cow)){
	unlink $cow || warn "Couldn't unlink COW file '$cow' : $!\n";
    }
    $me->{ubd}->[$dev] = undef;
}

1;

=head1 NAME

UML - class to control User-mode Linux

=head1 SYNOPSIS

 use UML;

 #################
 # class methods #
 #################
 $uml   = UML->new(kernel => $path_to_kernel,		# default "linux"
		   login_prompt => $login_prompt,       # "login:"
		   login => $account_name,              # "root"
		   password_prompt => $password_prompt, # "Password:"
		   password => $account_password,       # "root"
		   prompt => $shell_prompt_re,          # "darkstar:.*#"
		   halt => $halt_command);              # "halt"
 $uml->boot();
 $uml->command($command_string);
 $uml->halt();

 #######################
 # object data methods #
 #######################

 ########################
 # other object methods #
 ########################

=head1 DESCRIPTION

The UML class is used to control the execution of a user-mode kernel.
All of the arguments to UML::new are optional and will be defaulted if
not present.  The arguments and their values are as follows:
    kernel - the filename of the kernel executable
    arguments - a string containing the kernel command line
    login_prompt - a string matching the login prompt
    login - the account to log in to
    password_prompt - a string matching the password prompt
    password - the account's password
    prompt - a regular expression matching the shell prompt
    halt - the command used to halt the virtual machine

Once constructed, the UML object may be booted.  UML::boot() will
return after it has successfully logged in.

Then, UML::command may be called as many times as desired.  It will
return when the command has finished and the next shell prompt has
been seen.

When the testing is finished, UML::halt() is called to shut the
virtual machine down.
