# 
# Copyright (C) 2002 Jeff Dike (jdike@karaya.com)
# Licensed under the GPL
#

package UML::KConfig;

use strict;

my $config_index = 0;

my $built_pool = undef;
my @built = ( );

sub init_config {
    my $fh = shift;

    my @lines = <$fh>;
    close $fh;

    my $config = { options => { }, lines => [ ], config => undef };
    foreach my $line (@lines){
	my $option = undef;
	my $new = undef;
	if($line =~ /^CONFIG_([^=]*)=([^\n]*)\n$/){
	    $option = $1;
	    $new = [ $option, $2 ];
	}
	elsif($line =~ /^\# CONFIG_([^=]*) is not set\n$/){
	    $option = $1;
	    $new = [ $option, "n" ]
	    }
	else {
	    $new = $line;
	}

	push @{$config->{lines}}, $new;
	defined($option) and 
	    $config->{options}->{$option} = $#{$config->{lines}};
    }
    return($config);
}

sub change_config {
    my $config = shift;
    my $options = shift;

    foreach my $option (keys(%$options)){
	my $value = $options->{$option};
	ref($value) eq "CODE" and $value = $value->();

	!defined($config->{options}->{$option}) and 
	    die "Option '$option' not defined";
	my $n = $config->{options}->{$option};
	my $def = $config->{lines}->[$n];
	$def->[1] = $value;
    }
}

sub init_pool {
    my $pool = shift;
    my $have_linux = 0;
    my $linux_index = 0;

    my @kernels = glob("$pool/linux*");
    foreach my $kernel (@kernels){
	$kernel =~ /linux\.(\d*)$/ and 
	    $config_index = ($1 >= $config_index) ? $1 + 1 : $config_index;
	$kernel eq "$pool/linux" and $have_linux = 1;
    }
    if($have_linux){
	`mv $pool/linux $pool/linux.$config_index`;
	$? != 0 and die "Failed to move $pool/linux";
	$linux_index = $config_index;
	$config_index++;
    }
    foreach my $kernel (@kernels){
	$kernel eq "$pool/linux" and $kernel = "$pool/linux.$linux_index";
	if(!(open CONFIG, "$kernel --showconfig |")){
	    warn("Couldn't run '$kernel'\n");
	    next;
	}
	my $config = init_config(\*CONFIG);
	$config->{binary} = $kernel;

	push @built, $config;
    }
    $built_pool = $pool;
}

sub new {
    my $class = shift;
    my $pool = shift;
    my %requirements = @_;

    my $me = { options => { }, config => undef, pool => $pool };

    foreach my $option (keys(%requirements)){
	set_option($me, $option, $requirements{$option});
    }    

    bless($me, $class);

    return $me;
}

sub get {
    my $me = shift;
    my $option = shift;

    !defined($me->{options}->{$option}) and 
	die "Option '$option' doesn't exist";

    my $n = $me->{options}->{$option};
    my $val = $me->{lines}->[$n];
    my ($set, undef, $value) = @$val;
    $set == 0 and return(undef);
    return($value);
}

sub set_option {
    my $me = shift;
    my $option = shift;
    my $value = shift;

    $me->{options}->{$option} = $value;
}

sub preserve {
    my $me = shift;
    my $pool = $me->{pool};
    
    my $new = "$pool/linux.$config_index";
    my $output = `mv $pool/linux $new`;
    $? != 0 and die "Failed to rename UML kernel : '$output'";
    push @built, $me;

    $config_index++;
    $me->{config}->{binary} = $new;
}

sub build {
    my $me = shift;
    my $pool = $me->{pool};
    my $config = "$pool/.config";
    my $nprocs = int(`grep processor /proc/cpuinfo | wc -l`);
    my $make = "make";
    
    my $output = `cd $pool ; $make mrproper ARCH=um 2>&1`;
    $? != 0 and die "make mrproper failed : '$output'";

    open CONFIG, ">$config" or die "Couldn't open '$config' : $!";
    foreach my $line (@{$me->{config}->{lines}}){
	if(ref($line) eq ""){
	    print CONFIG $line;
	    next;
	}
	my ($option, $value) = @$line;
	print CONFIG "CONFIG_$option=$value\n";
    }
    close CONFIG;

    $output = `cd $pool ; $make oldconfig ARCH=um ; $make linux ARCH=um 2>&1`;
    $? != 0 and die "UML build failed : '$output'";

    $me->preserve();
}

sub produce {
    my $me = shift;

    !defined($built_pool) and init_pool($me->{pool});
    foreach my $config (@built){
	if(config_compatible($config, $me->{options})){
	    $me->{config} = $config;
	    last;
	}
    }

    if(!defined($me->{config})){
	my $pool = $me->{pool};
	my $defconfig = "$pool/arch/um/defconfig";
	open CONFIG, "<$defconfig" or die "Couldn't open '$defconfig' : $!";
	$me->{config} = init_config(\*CONFIG);
	change_config($me->{config}, $me->{options});
	$me->build();
    }
}

sub location {
    my $me = shift;

    !defined($me->{config}) and die "KConfig::location : UML not built";
    return($me->{config}->{binary});
}

sub config_compatible {
     my $config = shift;
     my $new = shift;

     foreach my $option (keys(%$new)){
	 my $new_val = $new->{$option};
	 my $n = $config->{options}->{$option};
	 my $old_val = $config->{lines}->[$n];

	 if(ref($new_val) eq "CODE"){
	     $new_val->($old_val->[1]) or return(0);
	 }
	 else {
	     $new_val ne $old_val->[1] and return(0);
	 }
     }
     return(1);

}

sub compatible {
     my $me = shift;
     my $new = shift;

     return(config_compatible($me->{config}, $new));
}

1;
