# 
# Copyright (C) 2002 Jeff Dike (jdike@karaya.com)
# Licensed under the GPL
#

package UML::TestSuite;

use strict;

sub new {
    my $class = shift;
    my $name = shift;
    my $cwd = `pwd`;
    my %tests = @_;

    chomp($cwd);
    push @INC, $cwd;
    my $me = { name => $name,
	       subsuites => { map { $_ => [ $tests{$_}, -1 ] } 
			      keys(%tests) } };
    
    bless($me, $class);
    return($me);
}

sub name {
    my $me = shift;

    return($me->{name});
}

# This is a bunch of trickery.  Each test that's loaded is given its own
# module.  This puts any procedures that it defines in the symbol table of
# that module.  Otherwise, those procedures end up in the symbol table of
# this module, and they will override any UML::TestSuite methods of the
# same name, which is very bad.
# So, the eval "package ..." stuff below does the new module for each test.
# $do is defined this way because the 'do $file' has to be done in the context
# of the test's module.  If the test calls into another module to get do_file,
# then the test's procedures will be loaded in the symbol table of that
# module.
# Defining $do as a string like this and expanding it in the eval causes
# each test to get its own private do_file, which loads the test's symbols
# in its own private symbol table.

my $do = 'sub do_file {
    my $file = shift;
    my $ret = do $file;

    if(!$ret){
	if($@){
	    die "Couldn\'t parse \'$file\' : $@\n";
	}
	elsif(!defined($ret)){
	    die "Couldn\'t run \'$file\': $!\n";
	}
	else {
	    die "Couldn\'t run \'$file\' for some unknown reason\n";
	}
    }
    return($ret);
}';

my $test_index = 0;

sub walk {
    my $suite = shift;
    my $root = shift;
    my $test = shift;
    my $config = shift;
    !defined($test) and $test = "*";

    if(-d "$root"){
	my ($base, $rest) = split(/\//, $test, 2);
	my @globbed = glob("$root/$base");
	foreach my $one (@globbed){
	    $one =~ s/$root\///;
	    my $name = $one;
	    $name =~ s/\.pl$//;
	    my $new;
	    defined($suite->{subsuites}->{$name}) and 
		$new = $suite->{subsuites}->{$name}->[0];
	    !defined($new) and $new = UML::TestSuite->new($name);
	    $new = walk($new, "$root/$one", $rest, $config->{$name});
	    defined($new) and $suite->{subsuites}->{$name} = [ $new, -1 ];
	}
	!%{$suite->{subsuites}} and $suite = undef;
	return($suite);
    }
    else {
	my $test;
	my ($name) = ($root =~ /.*\/([^\/]*)$/) ;
	my $file;

	if($root =~ /.*\.pl$/){
	    $file = $root;
	    $name =~ s/\.pl$//;
	}
	elsif(-f "$root.pl"){
	    $file = "$root.pl";
	}
	if(defined($file)){
	    $test = eval "package Test$test_index; $do; do_file('$file');";
	    $@ and warn($@);
	    $test_index++;
	}
	if(defined($test)){
	    $test->name($name);
	    return($test->config($config));
	}
    }
    return(undef);
}

sub populate {
    my $me = shift;
    my $root = shift;
    my $config = shift;

    !@_ and @_ = ( undef );
    foreach my $test (@_){
	walk($me, $root, $test, $config->{tests});
    }
    return($me);    
}

sub run_test {
    my ($test, $uml, $log, $name, $config, @after) = @_;

    eval { 
	$test->do($uml, $log, $name, $config);
	foreach my $post (@after){
	    $uml->command($post);
	}
    };
    if($@){
	$@ =~ s/\s*$//;
	return($@);
    }
    else {
	return(undef);
    }
}

my @configs = ();

my $port = 9000;

sub do {
    my $me = shift;
    my $config = shift;
    my $log = shift;
    my @results = ();
    
    $log->log_run("Starting tests : " . $me->name());

    my $start_time = time();
    foreach my $name (keys(%{$me->{subsuites}})){
	my $suite = $me->{subsuites}->{$name};
	my ($test, $index) = @$suite;
	if(ref($test) eq "UML::TestSuite"){
	    my $sub_config = $config->sub_test($name);
	    my (undef, undef, @subresults) = $test->do($sub_config, $log);
	    push @results, @subresults;
	}
	else {
	    my $name = $test->name();
	    $log->log_set("Test started: $name");
	    my $sub_config = $config->sub_test($name);

	    my $status = undef;
	    my $elapsed;
	    $config->add_defaults("con" => "port:$port", 
				  "ssl" => "port:$port");
	    $port++;
	    my ($kconfig, $uml, undef) = @{$configs[$index]};
	    if($test->require("state") eq "down"){
		$config->add_defaults("umid" => "temp$index");
		$configs[$index]->[2]--;
		my $uml = UML->new($config);
		$uml->args($config->uml_config(undef));
		$uml->args("kernel" => $kconfig->location());
		$uml->timer(sub { return($test->time_left()) } );
		my $start = time();
		$status = run_test($test, $uml, $log, $name, $sub_config);
		$elapsed = time() - $start;
	    }
	    else {
		$config->add_defaults("umid" => "perm$index");
		if(!defined($uml)){
		    $uml = UML->new();
		    $uml->args($config->uml_config(undef));
		    $uml->args("kernel" => $kconfig->location());
		    eval { $uml->boot($log->logger()); };
		    if($@){
			chomp $@;
			$log->log_set("Failed to boot : $@");
			die "Failed to boot : $@";
		    }
		    $configs[$index]->[1] = $uml;
		}
		$uml->command("rm -rf /tmp/test");
		$uml->command("mkdir /tmp/test");
		$uml->command("cd /tmp/test");
		$uml->timer(sub { return($test->time_left()) } );
		my $start = time();
		$status = run_test($test, $uml, $log, $name, $sub_config,
				   "cd ~", "rm -rf /tmp/test");
		$elapsed = time() - $start;
		$uml->timer(undef);
		if($status eq "Kernel panic"){
		    eval { $uml->boot($log->logger()); };
		    if($@){
			chomp $@;
			$log->log_set("Failed to boot : $@");
			die "Failed to boot : $@";
		    }		    
		}
		if(--$configs[$index]->[2] == 0){
		    $uml->halt();
		    $configs[$index]->[1] = undef;
		}
	    }
	    if(defined($status)){
		$log->log_test("Test failed : $status");
		push @results, [ $name, 0, $elapsed, $status ];
	    }
	    else {
		push @results, [ $name, 1, $elapsed  ];
	    }
	    $log->log_set("Test finished: $name");
	}
    }
    my $end_time = time();

    $log->log_run("Finished tests : " . $me->name());
    return($start_time, $end_time, @results);
}

sub compatible_uml {
    my $need = shift;

    foreach my $i (0 .. $#configs){
	my ($have, undef, undef) = @{$configs[$i]};
	$have->compatible($need) and return($i);
    }
    return(-1);
}

sub build_umls {
    my $me = shift;
    my $config = shift;

    foreach my $name (keys(%{$me->{subsuites}})){
	my $suite = $me->{subsuites}->{$name};
	my ($test, undef) = @$suite;
	if(ref($test) eq "UML::TestSuite"){
	    $test->build_umls($config);
	}
	else {
	    my %needs = %{$test->kconfig()};
	    my $n = compatible_uml(\%needs);
	    if($n == -1){
		my $pool = $config->get_value("kernel_pool");
		!defined($pool) and 
		    die "build_umls : kernel_pool not defined";
		my $new = UML::KConfig->new($pool, %needs);
		$new->produce();
		push @configs, [ $new, undef, 0, $config->clone() ];
		$n = $#configs;
	    }
	    $suite->[1] = $n;
	    $configs[$n]->[2]++;
	    $test->config_uml($configs[$n]->[3]);
	}
    }
}

sub run {
    my $me = shift;
    my $config = shift;
    my $log = shift;

    $log->log_run("Building UMLs");
    $me->build_umls($config);
    $log->log_run("Done");

    $me->do($config, $log);
}

1;
